import { normalizePath } from "vite";
import glob from "fast-glob";
import * as fs from "node:fs/promises";
import * as fsSync from "node:fs";
import * as path from "node:path";
import { bold, cyan } from "kleur/colors";
import { info, warn } from "../core/logger/core.js";
import { appendForwardSlash, prependForwardSlash } from "../core/path.js";
import { contentFileExts, CONTENT_FLAG, VIRTUAL_MODULE_ID } from "./consts.js";
import { escapeViteEnvReferences } from "../vite-plugin-utils/index.js";
import { pathToFileURL } from "node:url";
import {
  getEntryData,
  getEntrySlug,
  loadContentConfig,
  NotFoundError,
  parseFrontmatter
} from "./utils.js";
import * as devalue from "devalue";
const CONTENT_BASE = "types.generated";
const CONTENT_FILE = CONTENT_BASE + ".mjs";
const CONTENT_TYPES_FILE = CONTENT_BASE + ".d.ts";
const msg = {
  collectionAdded: (collection) => `${cyan(collection)} collection added`,
  entryAdded: (entry, collection) => `${cyan(entry)} added to ${bold(collection)}.`
};
function astroContentVirtualModPlugin({ settings }) {
  const paths = getPaths({ srcDir: settings.config.srcDir });
  const relContentDir = appendForwardSlash(
    prependForwardSlash(path.relative(settings.config.root.pathname, paths.contentDir.pathname))
  );
  const entryGlob = relContentDir + "**/*.{md,mdx}";
  const astroContentModContents = fsSync.readFileSync(new URL(CONTENT_FILE, paths.generatedInputDir), "utf-8").replace("@@CONTENT_DIR@@", relContentDir).replace("@@ENTRY_GLOB_PATH@@", entryGlob).replace("@@RENDER_ENTRY_GLOB_PATH@@", entryGlob);
  const astroContentVirtualModuleId = "\0" + VIRTUAL_MODULE_ID;
  return {
    name: "astro-content-virtual-mod-plugin",
    enforce: "pre",
    resolveId(id) {
      if (id === VIRTUAL_MODULE_ID) {
        return astroContentVirtualModuleId;
      }
    },
    load(id) {
      if (id === astroContentVirtualModuleId) {
        return {
          code: astroContentModContents
        };
      }
    }
  };
}
function astroContentServerPlugin({
  settings,
  contentConfig,
  logging
}) {
  const paths = getPaths({ srcDir: settings.config.srcDir });
  let contentDirExists = false;
  let contentGenerator;
  async function createContentGenerator() {
    const contentTypes = {};
    let events = [];
    let debounceTimeout;
    let eventsSettled;
    const contentTypesBase = await fs.readFile(
      new URL(CONTENT_TYPES_FILE, paths.generatedInputDir),
      "utf-8"
    );
    async function init() {
      const pattern = new URL("./**/", paths.contentDir).pathname + "*.{md,mdx}";
      const entries = await glob(pattern);
      for (const entry of entries) {
        queueEvent({ name: "add", entry }, { shouldLog: false });
      }
      await eventsSettled;
    }
    async function onEvent(event, opts) {
      const shouldLog = (opts == null ? void 0 : opts.shouldLog) ?? true;
      if (event.name === "addDir" || event.name === "unlinkDir") {
        const collection = path.relative(paths.contentDir.pathname, event.entry);
        const isCollectionEvent = collection.split(path.sep).length === 1;
        if (!isCollectionEvent)
          return;
        switch (event.name) {
          case "addDir":
            addCollection(contentTypes, JSON.stringify(collection));
            if (shouldLog) {
              info(logging, "content", msg.collectionAdded(collection));
            }
            break;
          case "unlinkDir":
            removeCollection(contentTypes, JSON.stringify(collection));
            break;
        }
      } else {
        const fileType = getEntryType(event.entry, paths);
        if (fileType === "config") {
          contentConfig = await loadContentConfig({ settings });
          return;
        }
        if (fileType === "unknown") {
          warn(
            logging,
            "content",
            `${cyan(
              path.relative(paths.contentDir.pathname, event.entry)
            )} is not a supported file type. Skipping.`
          );
          return;
        }
        const entryInfo = getEntryInfo({ entryPath: event.entry, contentDir: paths.contentDir });
        if (entryInfo instanceof Error)
          return;
        const { id, slug, collection } = entryInfo;
        const collectionKey = JSON.stringify(collection);
        const entryKey = JSON.stringify(id);
        const collectionConfig = contentConfig instanceof Error ? void 0 : contentConfig.collections[collection];
        switch (event.name) {
          case "add":
            if (!(collectionKey in contentTypes)) {
              addCollection(contentTypes, collectionKey);
            }
            if (!(entryKey in contentTypes[collectionKey])) {
              addEntry(contentTypes, collectionKey, entryKey, slug, collectionConfig);
            }
            if (shouldLog) {
              info(logging, "content", msg.entryAdded(entryInfo.slug, entryInfo.collection));
            }
            break;
          case "unlink":
            if (collectionKey in contentTypes && entryKey in contentTypes[collectionKey]) {
              removeEntry(contentTypes, collectionKey, entryKey);
            }
            break;
          case "change":
            break;
        }
      }
    }
    function queueEvent(event, eventOpts) {
      if (!event.entry.startsWith(paths.contentDir.pathname))
        return;
      if (event.entry.endsWith(CONTENT_TYPES_FILE))
        return;
      events.push(onEvent(event, eventOpts));
      runEventsDebounced();
    }
    function runEventsDebounced() {
      eventsSettled = new Promise((resolve, reject) => {
        try {
          debounceTimeout && clearTimeout(debounceTimeout);
          debounceTimeout = setTimeout(async () => {
            await Promise.all(events);
            await writeContentFiles({
              contentTypes,
              paths,
              contentTypesBase,
              hasContentConfig: !(contentConfig instanceof NotFoundError)
            });
            resolve();
          }, 50);
        } catch (e) {
          reject(e);
        }
      });
    }
    return { init, queueEvent };
  }
  return [
    {
      name: "content-flag-plugin",
      async load(id) {
        const { pathname, searchParams } = new URL(id, "file://");
        if (isContentFlagImport({ pathname, searchParams })) {
          const rawContents = await fs.readFile(pathname, "utf-8");
          const {
            content: body,
            data: unparsedData,
            matter: rawData
          } = parseFrontmatter(rawContents, pathname);
          const entryInfo = getEntryInfo({ entryPath: pathname, contentDir: paths.contentDir });
          if (entryInfo instanceof Error)
            return;
          const _internal = { filePath: pathname, rawData };
          const partialEntry = { data: unparsedData, body, _internal, ...entryInfo };
          const collectionConfig = contentConfig instanceof Error ? void 0 : contentConfig.collections[entryInfo.collection];
          const data = collectionConfig ? await getEntryData(partialEntry, collectionConfig) : unparsedData;
          const slug = collectionConfig ? await getEntrySlug({ ...partialEntry, data }, collectionConfig) : entryInfo.slug;
          const code = escapeViteEnvReferences(`
export const id = ${JSON.stringify(entryInfo.id)};
export const collection = ${JSON.stringify(entryInfo.collection)};
export const slug = ${JSON.stringify(slug)};
export const body = ${JSON.stringify(body)};
export const data = ${devalue.uneval(data)};
export const _internal = {
	filePath: ${JSON.stringify(pathname)},
	rawData: ${JSON.stringify(rawData)},
};
`);
          return { code };
        }
      },
      async transform(code, id) {
        if (isContentFlagImport(new URL(id, "file://"))) {
          return { code: escapeViteEnvReferences(code) };
        }
      }
    },
    {
      name: "astro-content-server-plugin",
      async config() {
        try {
          await fs.stat(paths.contentDir);
          contentDirExists = true;
        } catch {
          return;
        }
        info(logging, "content", "Generating entries...");
        contentGenerator = await createContentGenerator();
        await contentGenerator.init();
      },
      async configureServer(viteServer) {
        if (contentDirExists) {
          info(
            logging,
            "content",
            `Watching ${cyan(
              paths.contentDir.href.replace(settings.config.root.href, "")
            )} for changes`
          );
          attachListeners();
        } else {
          viteServer.watcher.on("addDir", (dir) => {
            if (dir === paths.contentDir.pathname) {
              info(logging, "content", `Content dir found. Watching for changes`);
              contentDirExists = true;
              attachListeners();
            }
          });
        }
        function attachListeners() {
          viteServer.watcher.on("all", async (event, entry) => {
            if (["add", "unlink", "change"].includes(event) && getEntryType(entry, paths) === "config") {
              for (const modUrl of viteServer.moduleGraph.urlToModuleMap.keys()) {
                if (isContentFlagImport(new URL(modUrl, "file://"))) {
                  const mod = await viteServer.moduleGraph.getModuleByUrl(modUrl);
                  if (mod) {
                    viteServer.moduleGraph.invalidateModule(mod);
                  }
                }
              }
            }
          });
          viteServer.watcher.on("add", (entry) => {
            contentGenerator.queueEvent({ name: "add", entry });
          });
          viteServer.watcher.on(
            "addDir",
            (entry) => contentGenerator.queueEvent({ name: "addDir", entry })
          );
          viteServer.watcher.on(
            "change",
            (entry) => contentGenerator.queueEvent({ name: "change", entry })
          );
          viteServer.watcher.on("unlink", (entry) => {
            contentGenerator.queueEvent({ name: "unlink", entry });
          });
          viteServer.watcher.on(
            "unlinkDir",
            (entry) => contentGenerator.queueEvent({ name: "unlinkDir", entry })
          );
        }
      }
    }
  ];
}
function getPaths({ srcDir }) {
  return {
    cacheDir: new URL("./content/", srcDir),
    contentDir: new URL("./content/", srcDir),
    generatedInputDir: new URL("../../src/content/template/", import.meta.url),
    config: new URL("./content/config", srcDir)
  };
}
function isContentFlagImport({ searchParams, pathname }) {
  return searchParams.has(CONTENT_FLAG) && contentFileExts.some((ext) => pathname.endsWith(ext));
}
function addCollection(contentMap, collectionKey) {
  contentMap[collectionKey] = {};
}
function removeCollection(contentMap, collectionKey) {
  delete contentMap[collectionKey];
}
function addEntry(contentTypes, collectionKey, entryKey, slug, collectionConfig) {
  const dataType = (collectionConfig == null ? void 0 : collectionConfig.schema) ? `InferEntrySchema<${collectionKey}>` : "any";
  const slugType = (collectionConfig == null ? void 0 : collectionConfig.slug) ? "string" : JSON.stringify(slug);
  contentTypes[collectionKey][entryKey] = `{
  id: ${entryKey},
  slug: ${slugType},
  body: string,
  collection: ${collectionKey},
  data: ${dataType}
}`;
}
function removeEntry(contentTypes, collectionKey, entryKey) {
  delete contentTypes[collectionKey][entryKey];
}
function getEntryInfo({
  entryPath,
  contentDir
}) {
  const relativeEntryPath = normalizePath(path.relative(contentDir.pathname, entryPath));
  const collection = path.dirname(relativeEntryPath).split(path.sep).shift();
  if (!collection)
    return new Error();
  const id = path.relative(collection, relativeEntryPath);
  const slug = id.replace(path.extname(id), "");
  return {
    id,
    slug,
    collection
  };
}
function getEntryType(entryPath, paths) {
  const { dir, ext, name } = path.parse(entryPath);
  const { pathname } = new URL(name, appendForwardSlash(pathToFileURL(dir).href));
  if ([".md", ".mdx"].includes(ext)) {
    return "content";
  } else if (pathname === paths.config.pathname) {
    return "config";
  } else {
    return "unknown";
  }
}
async function writeContentFiles({
  paths,
  contentTypes,
  contentTypesBase,
  hasContentConfig
}) {
  let contentTypesStr = "";
  for (const collectionKey in contentTypes) {
    contentTypesStr += `${collectionKey}: {
`;
    for (const entryKey in contentTypes[collectionKey]) {
      const entry = contentTypes[collectionKey][entryKey];
      contentTypesStr += `${entryKey}: ${entry},
`;
    }
    contentTypesStr += `},
`;
  }
  contentTypesBase = contentTypesBase.replace("// @@ENTRY_MAP@@", contentTypesStr);
  contentTypesBase = contentTypesBase.replace(
    "'@@CONTENT_CONFIG_TYPE@@'",
    hasContentConfig ? `typeof import(${JSON.stringify(paths.config.pathname)})` : "never"
  );
  try {
    await fs.stat(paths.cacheDir);
  } catch {
    await fs.mkdir(paths.cacheDir);
  }
  await fs.writeFile(new URL(CONTENT_TYPES_FILE, paths.cacheDir), contentTypesBase);
}
export {
  astroContentServerPlugin,
  astroContentVirtualModPlugin,
  getPaths
};
